##------------------##
## Python libraries ##
##------------------##

import sys
import numpy as np

##-----------##
## ML-agents ##
##-----------##

from mlagents_envs.environment import ActionTuple, UnityEnvironment
import EnvChannel
import TaskSideChannel
from Node import Node

##----##
## TP ##
##----##

from tp_system import System as TP_System
from ta_state import Location
from tp_task import Task



simulation_going = True

# Load the side channels and Unity environment.
envChannel = EnvChannel.EnvChannel()
taskChannel = TaskSideChannel.TaskSideChannel()
env = UnityEnvironment(file_name=None, side_channels=[envChannel, taskChannel])  # Open unity and press play to execute
env.reset() # reset the environment and get info about environment through envChannel

# Specify the used behaviour name for steering the agent
behavior_names = list(env.behavior_specs.keys())
behavior_name = behavior_names[0]

# info obtained from environment
environment_info = envChannel.EnvData
# set up links between task side channel and environment info
environment_info.task_side_channel = taskChannel
taskChannel.environment_data = environment_info

# Retrieve information from environment
matrix = environment_info.matrix
number_of_robots = environment_info.number_of_robots
rotation = environment_info.rotation
observation_size = environment_info.observation_size
action_size = environment_info.action_size
robot_can_go_backwards = environment_info.can_go_backwards

# Initialize system
system = TP_System(matrix, rotation)

# Callback called when an agent assigns itself a task
def on_task_assigned(agent_id, task_id):
    environment_info.assign_task(agent_id, system.get_task(task_id))

##----------------##
## Robot Creation ##
##----------------##

# Number of agents in the configuration file
n_agents = len(environment_info.instance_ids)

# Create all the agents 
for robot_id, start_node in environment_info.current_nodes_dictionary.items():
    # Convert the Node that represents the agent's location to a Location
    robot_position = Location(start_node.y, start_node.x, start_node.g)
    # Create agent    
    system.add_agent(robot_id, robot_position, on_task_assigned)

# Each agent requests 1 action in each timestep
# next timestep starts when all agents have requested an action in this timestep
actions_requested_in_timestep = 0
# Keep track of current timestep
timestep = 0

while simulation_going:
    # If there are new tasks that need to be assigned add them to the task set
    if len(environment_info.tasks_to_be_assigned) != 0:
        for task in environment_info.tasks_to_be_assigned:
            # Add task to task set
            system.add_task(task)
            # Task is already in task set, remove it from this list to avoid duplicates in task set
            environment_info.tasks_to_be_assigned.remove(task)

    # Get information on the agents that request an action
    decision_steps, terminal_steps = env.get_steps(behavior_name)
    agents = decision_steps.agent_id
    observations = decision_steps.obs

    # Actions that will be sent back to Unity
    actions = []
    action_tuple = ActionTuple()  # actions to send to unity

    # Create an action for all agents that request an action
    for i in range(len(agents)):
        
        ##--------------##
        ## OBSERVATIONS ##
        ##--------------## 

        # Keep track of how many agents have requested an action in this timestep
        actions_requested_in_timestep += 1
        
        # Collect agent's observations        
        observation = observations[0][i]
        # Extract observations
        robot_id_unity = observation[0] # robot id in unity used to get correct precomputed path
        robot_pos_x = observation[2]
        robot_pos_y = observation[1]
        robot_rot = observation[3]
        # Convert coordinates to the agent's current position
        robot_position = Location(robot_pos_x, robot_pos_y, robot_rot)

        # Get Python Agent that represents robot
        robot = system.get_agent(int(robot_id_unity))

        ##-------------##
        ## FOLLOW PATH ##
        ##-------------##

        # Follow previously computed path or plan new path when at the end of path
        next_position = robot.move(system.token, timestep)
        # Convert Location to Node
        next_action = Node(next_position.row, next_position.col, next_position.rotation)
        # Check if a task needs to be picked up / delivered
        pickup_action = 0
        delivery_action = 0
        if next_position.to_pickup:
            # Agent has to pick up his task
            pickup_action = 1
        elif next_position.to_deliver:
            # Agent has to deliver his task
            delivery_action = 1
            # Agent has no task assigned anymore as the task is delivered
            system.token.assignments[robot_id_unity] = None

        # Send action to agent in Unity
        actions.extend([robot_id_unity,next_action.x, next_action.y, next_action.g, pickup_action, delivery_action])              

    # When all agents have moved in this timestep update the timestep
    if (actions_requested_in_timestep == n_agents):
        # Update the timestep
        timestep += 1
        # Reset the number of agents that have requested a timestep in this timestep
        actions_requested_in_timestep = 0
    
    # Send actions to Unity
    actions = np.array(actions)
    actions.resize(len(agents), action_size)
    action_tuple.add_discrete(actions)

    env.set_actions(behavior_name, action_tuple)
    env.step()






