"""
Created on Wed Apr 22 07:43:03 2020

@author: Pieter Cawood

"""
from heapq import *
from ta_state import *

def get_heuristic(start_location, goal_location, rotation):
    # Admissible heuristic - manhattan distance
    return abs(start_location.col - goal_location.col) + abs(start_location.row - goal_location.row)

class AStar:
    def __init__(self, world, rotation):
        self.rotation = rotation
        self.world = world
        self.dim_x = len(world[0])
        self.dim_y = len(world)
        self.obstacles = {}
        self.path_ends = dict()
        self.other_agent_paths = []

    def state_possible(self, current_state, new_state):
        # Check if wall
        not_obstacle = False
        if new_state.location not in self.obstacles:
            not_obstacle = True
        # Check if location is in dimensions of grid and does not collide with another location/wall
        return 0 <= new_state.location.col < self.dim_x and 0 <= new_state.location.row < self.dim_y \
               and not_obstacle and not self.is_collision(current_state, new_state)

    def is_collision(self, current_state, new_state):
        for other_agent_path in self.other_agent_paths:
            # Vertex collision
            if new_state.time in other_agent_path:
                if other_agent_path[new_state.time].row == new_state.location.row and other_agent_path[new_state.time].col == new_state.location.col:
                    return True
            # Edge collision
            if new_state.time in other_agent_path and current_state.time in other_agent_path:
                if other_agent_path[new_state.time].row == current_state.location.row and \
                    other_agent_path[new_state.time].col == current_state.location.col and \
                        other_agent_path[current_state.time].row == new_state.location.row and \
                            other_agent_path[current_state.time].col == new_state.location.col:
                    return True
            # Last position of agent is his parking location
            if new_state.location in self.path_ends:
                busy_time = self.path_ends[new_state.location]
                if new_state.time >= busy_time:
                    return True
        return False

    def get_neighbours(self, current_state):
        neighbours = []
        neighbour_time_step = current_state.time + 1
        # Wait in state
        new_state = State(neighbour_time_step, Location(current_state.location.col, current_state.location.row, current_state.location.rotation))
        if self.state_possible(current_state, new_state):
            neighbours.append(new_state)
        # Add rotation
        new_state = State(neighbour_time_step, Location(current_state.location.col, current_state.location.row, (current_state.location.rotation+self.rotation) % 360))
        if self.state_possible(current_state, new_state):
            neighbours.append(new_state)
        # Remove rotation
        new_state = State(neighbour_time_step, Location(current_state.location.col, current_state.location.row, (current_state.location.rotation-self.rotation) % 360))
        if self.state_possible(current_state, new_state):
            neighbours.append(new_state)
        if current_state.location.rotation == 0:
            # Move down
            new_state = State(neighbour_time_step, Location(current_state.location.col, current_state.location.row - 1, current_state.location.rotation))
            if self.state_possible(current_state, new_state):
                neighbours.append(new_state)
        elif current_state.location.rotation == 90:
            # Move right
            new_state = State(neighbour_time_step, Location(current_state.location.col + 1, current_state.location.row, current_state.location.rotation))
            if self.state_possible(current_state, new_state):
                neighbours.append(new_state)
        elif current_state.location.rotation == 180:
            # Move up
            new_state = State(neighbour_time_step, Location(current_state.location.col, current_state.location.row + 1, current_state.location.rotation))
            if self.state_possible(current_state, new_state):
                neighbours.append(new_state)
        elif current_state.location.rotation == 270:
            # Move left
            new_state = State(neighbour_time_step, Location(current_state.location.col - 1, current_state.location.row, current_state.location.rotation))
            if self.state_possible(current_state, new_state):
                neighbours.append(new_state)       
        return neighbours

    def create_obstacles(self, other_agent_paths):
        n_rotations = int(360/self.rotation)
        # Keep track of all walls in the environment
        for y in range(self.dim_y):
            for x in range(self.dim_x):
                # 1 represent a wall in the environment
                if self.world[y][x] == 1:
                    # A wall is always a collision, no matter the rotation
                    for i in range(n_rotations):
                        rot = i * self.rotation
                        self.obstacles[Location(x, y, rot)] = True
        # Keep track of where the paths of the other agents end
        self.path_ends = dict()
        for path in other_agent_paths:
            # Get the last timestep in the path of the agent
            last_time = list(path)[-1]
            # Get the last location in the path of the agent
            last_pos = path[last_time]
            # Two agents always collide when they share the same location, no matter the rotation
            for i in range(n_rotations):
                rot = i * self.rotation
                self.path_ends[Location(last_pos.col, last_pos.row, rot)] = last_time
        # Keep track of the other paths of agents to detect collisions
        self.other_agent_paths = other_agent_paths

    # Returns the latest timestep at which a specific location will be used
    def find_lastest_use(self, location):
        # Keep track of all the uses of that location
        uses = []
        # Check in the path of each agent if that agent uses that location, if yes, add it to the list
        for agent_path in self.other_agent_paths:
            for t, loc in agent_path.items():
                if loc.col == location.col and loc.row == location.row:
                    uses.append(t)
        # If the location is used at least once, return the maximum
        if uses:
            return max(uses)
        # If the location is never used, return 0
        else:
            return 0
        

    def search(self, time_step, start_location, goal_location, other_agent_paths,
               release_time=0, find_step = 0):
        self.create_obstacles(other_agent_paths)
        initial_state = State(time_step, start_location)
        # Find the lastest use of the goal location to avoid parking on a location that is still going to be used
        lastest_use = self.find_lastest_use(goal_location)
        came_from = {}
        open_list = [initial_state]
        closed_set = set()
        initial_state.g_score = 0
        initial_state.f_score = get_heuristic(start_location, goal_location, self.rotation)
        while open_list:
            current = heappop(open_list)
            if current.location == goal_location:
                if current in came_from and came_from[current].location == goal_location:
                    if current.time > lastest_use:
                        new_path = [current]
                        while current in came_from:
                            current = came_from[current]
                            new_path.append(current)
                        # Reverse path
                        return new_path[::-1]
            closed_set |= {current}
            neighbours = self.get_neighbours(current)
            neighbour: State
            for neighbour in neighbours:
                # Only explore new states
                if (neighbour not in closed_set) and (neighbour not in open_list):
                    came_from[neighbour] = current
                    neighbour.g = current.g + 1
                    neighbour.f = neighbour.g + get_heuristic(neighbour.location, goal_location, self.rotation)
                    heappush(open_list, neighbour)
        return False
