# Python modules
import sys
import numpy as np
# Pathfinding modules
from ta_astar import AStar
from ta_state import Location, h


class Agent:
    def __init__(self, id, initialLocation, system, environment, rotation, on_task_assigned):
        self.id = id
        # Position
        self.pos = initialLocation
        # Plan paths
        self.a_star = AStar(environment, rotation)
        # Used to request token when looking for a task
        self.system = system

        self.on_task_assigned = on_task_assigned

    def look_for_task(self, token, timestep):
        availableTasks = [] # List of all the tasks that are available to the agent

        ##-- Check if pickup/delivery is not end of path of any other agent --##
        # Get last locations of all paths in token
        pathEnds = token.get_pathends(self.id)
        # If pickup/delivery is not in pathEnds it's an available task
        for t in token.taskSet:
            if t.pickup not in pathEnds and t.delivery not in pathEnds:
                availableTasks.append(t)
        # If there is no available task plan a path to a location where the agent can wait
        if len(availableTasks) == 0:
            self.path2(token, timestep)
        # There is at least one task the agent can execute
        else:
            # Find the "best" task (best = minimum h-value)
            h_scores = np.array(map(lambda t: h(self.pos, t.pickup), availableTasks))
            min_task = availableTasks[np.argmin(h_scores)]
            # Plan path for that task
            self.path1(token, min_task, timestep)
        

    def path1(self, token, task, timestep):
        """[summary]
        Args:
            token (Token): The global token provided by the system
            task (Task): The task for which a path needs to be planned
            timestep (int): Current timestep
        """

        # All needed locations for the pathplanning
        current_location = Location(self.pos.col, self.pos.row, self.pos.rotation) 
        pickup_location = Location(task.pickup.col, task.pickup.row, task.pickup.rotation)
        delivery_location = Location(task.delivery.col, task.delivery.row, task.delivery.rotation)
        
        ##----------------##
        ## Path to pickup ##
        ##----------------##

        # Get the paths of all the other agents
        other_agent_paths = token.get_other_agents_paths(timestep, self.id)
        # Plan a path from the current location of the agent to the pickup location of the task
        pickup_path = self.a_star.search(timestep, current_location, pickup_location, other_agent_paths)
        # There should always be a path if the problem is well formed
        if not pickup_path:
            raise ValueError("[TP] No path to pickup location - Problem not well formed")
        # Add newly planned path to the agent's path in the token
        for state in pickup_path:
            token.paths[self.id][state.time] = state.location
        old_timestep = timestep

        # Update the timestep to the timestep at the end of the planned path 
        timestep = timestep + len(pickup_path) - 1

        timestep_at_pickup = timestep

        ##------------------##
        ## Path to delivery ##
        ##------------------##

        # Get the paths of all the other agents
        other_agent_paths = token.get_other_agents_paths(timestep, self.id)
        # Plan a path from the pickup location of the task to the delivery location of the task
        delivery_path = self.a_star.search(timestep, pickup_location, delivery_location, other_agent_paths)
        # There should always be a path if the problem is well formed
        if not delivery_path:
            raise ValueError("[TP] No path to delivery location - Problem not well formed")
        # Add newly planned path to the agent's path in the token
        for state in delivery_path:
            token.paths[self.id][state.time] = state.location

        timestep = timestep + len(delivery_path) - 1

        ##---------------------------##
        ## Pickup / Delivery actions ##
        ##---------------------------##

        # Register in the last location of the path to the pickup location that the agent needs to pickup a task and register the task ID
        token.paths[self.id][timestep_at_pickup].to_pickup = True
        token.paths[self.id][timestep_at_pickup].task_id = task.id

        # Register in the last location of the path to the delivery location that the agent needs to pickup a task and register the task ID
        token.paths[self.id][timestep].to_deliver = True
        token.paths[self.id][timestep].task_id = task.id

        # Assign task to agent
        token.assignments[self.id] = task.id
        # Remove the task from the taskset
        token.taskSet.remove(task)

    def path2(self, token, timestep):
        ##-- Check if current location is not delivery location of any task in taskSet --##
        # Indicates whether agent needs to move
        needToMove = False
        for t in token.taskSet:
            # If the agent is parked on the delivery location of a task he needs to move 
            if t.delivery == self.pos:
                needToMove = True
        if not needToMove:
            # Agent doesn't need to move, update the path with a trivial path to current position
            token.paths[self.id][timestep] = Location(self.pos.col, self.pos.row, self.pos.rotation)
        else:
            ##-- Find a free endpoint to move to --##
            endpoints = self.system.get_free_endpoints()
            # Select the best endpoint (best = smallest h-score)
            h_scores = list(map(lambda e: h(self.pos, e), endpoints))
            chosen_endpoint = endpoints[np.argmin(h_scores)]

            ##-- Plan path to free endpoint --##
            # Make sure planned path doesn't collide with already planned paths
            current_location = Location(self.pos.col, self.pos.row, self.pos.rotation) 
            endpoint_location = Location(chosen_endpoint.col, chosen_endpoint.row, chosen_endpoint.rotation)
            # Get the paths of all the other agents
            other_agent_paths = token.get_other_agents_paths(timestep, self.id)
            # Plan a path from the agent's current location to the chosen endpoint
            path_to_endpoint = self.a_star.search(timestep, current_location, endpoint_location, other_agent_paths)
            # There should always be a path if the problem is well formed
            if not path_to_endpoint:
                raise ValueError("[TP] No path to chosen endpoint - Problem not well formed")
            # Add newly planned path to the agent's path in the token
            for state in path_to_endpoint:
                token.paths[self.id][state.time] = state.location

    def move(self, token, timestep):
        # Get the current path of the agent
        path = token.paths[self.id]
        # Check if the agent has already planned a move for this timestep
        if timestep not in path:
            # If agent has not yet planned a move for this timestep it requests the token and looks for a task
            # Start looking for a new task
            self.look_for_task(token, timestep)
            # Check if the agent has assigned a task to itself and notify Unity if that is the case (using callback)
            if token.assignments[self.id] is not None:
                self.on_task_assigned(self.id, token.assignments[self.id])

        # Select the move of the agent in this timestep
        next_move = token.paths[self.id][timestep]
        # Update the agent's position
        self.pos = next_move
        # Return the agent's next position            
        return next_move
