# Python modules
import sys
import numpy as np
# Pathfinding modules
from ta_state import Location
# Own modules
from tp_agent import Agent
from tp_token import Token

# System manages the token
class System:
    def __init__(self, env, rotation):
        # Keep track of the warehouse setup
        self.environment = env
        # Keep track of rotation
        self.rotation = rotation

        # Keep track of all agents in the system
        self.agents = dict()
        # Keep track of all tasks
        self.tasks = dict()
        # Token setup
        self.token = Token()
        # Keep track of all the non-task/task endpoints
        self.task_endpoints = []
        self.non_task_endpoints = []
        # Get all the endpoints from the environment
        self.init_endpoints()

    # Find all the task and non-task endpoints in the environment
    def init_endpoints(self):
        # Get the dimensions of the environment
        dim_y = len(self.environment)
        dim_x = len(self.environment[0])
        for y in range(dim_y):
            for x in range(dim_x):
                # Non-task endpoints (parking locations) are represented by 2
                if self.environment[y][x] == 2:
                    self.non_task_endpoints.append(Location(x, y, 0))
                # Task endpoints (pickup/delivery locations) are represented by 3
                elif self.environment[y][x] == 3:
                    self.task_endpoints.append(Location(x, y, 0))

    ##---------##
    ## Getters ##
    ##---------##
        
    def get_agent(self, a_id):
        return self.agents[a_id]
    
    def get_task(self, task_id):
        return self.tasks[task_id]

    def get_endpoints(self):
        return np.concatenate((self.non_task_endpoints, self.task_endpoints))

    def get_non_task_endpoints(self):
        return self.non_task_endpoints

    def get_free_endpoints(self):
        # List of all the endpoints
        all_endpoints = self.get_endpoints()
        free_endpoints = []
        # Get a list of the last positions of each agent's path
        endPaths = self.token.get_pathends()
        # Generate a list of all the delivery locations of the tasks in the taskSet
        deliveryLocations = []
        for t in self.token.taskSet:
            deliveryLocations.append(t.delivery)
        for e in all_endpoints:
            # If endpoint is not the end of the path of any agent
            # or is not the delivery location of any task it's free
            if e not in endPaths and e not in deliveryLocations:
                free_endpoints.append(e)
        return free_endpoints

    ##---------------------##
    ## Adding new elements ##
    ##---------------------##

    def add_task(self, task):
        self.token.add_task(task)
        self.tasks[task.id] = task
        return True
        
    def add_agent(self, agent_id, agent_pos, on_task_assigned):
        agent = Agent(agent_id, agent_pos, self, self.environment, self.rotation, on_task_assigned)
        self.agents[agent_id] = agent
        self.token.add_agent(agent)

    ##-------------------##
    ## Utilitary methods ##
    ##-------------------##

    def request_token(self):
        return self.token

    def is_endpoint(self, pos):
        if pos in self.non_task_endpoints or pos in self.task_endpoints:
            return True
        else:
            return False
        

